__all__ = [
    "RakPeer",
    "Packet",
    "PacketPriority",
    "PacketReliability",
    "MessageIdentifiers",
    "ConnectionAttemptError",
    "StartupError",
]

for name in __all__:
    from importlib import import_module

    module = import_module("raknet.raknet_python")
    globals()[f"_{name}"] = module.__dict__[name]
    del module

RakPeer = globals()["_RakPeer"]
Packet = globals()["_Packet"]
PacketPriority = globals()["_PacketPriority"]
PacketReliability = globals()["_PacketReliability"]
MessageIdentifiers = globals()["_MessageIdentifiers"]
ConnectionAttemptError = globals()["_ConnectionAttemptError"]
StartupError = globals()["_StartupError"]



import concurrent.futures
import time

from raknet import RakPeer, MessageIdentifiers, PacketPriority, PacketReliability

ID_GAME_MESSAGE_1 = MessageIdentifiers.ID_USER_PACKET_ENUM + 1


def run_server(server: RakPeer):
    timeout = time.time() + 5
    while time.time() < timeout:
        packet = server.receive()
        if packet is None:
            continue
        if packet.data[0] == MessageIdentifiers.ID_NEW_INCOMING_CONNECTION:
            print("A connection is incoming.")
        elif packet.data[0] == MessageIdentifiers.ID_DISCONNECTION_NOTIFICATION:
            print("A client has disconnected.")
        elif packet.data[0] == MessageIdentifiers.ID_CONNECTION_LOST:
            print("A client lost the connection.")
        elif packet.data[0] == ID_GAME_MESSAGE_1:
            print(packet.data[1:], packet.system_address)
            assert packet.data[1:] == b"Hello, World"
            return

    assert False, "Game message was not received"


def run_client(client: RakPeer):
    timeout = time.time() + 5
    while time.time() < timeout:
        packet = client.receive()
        if packet is None:
            continue
        if packet.data[0] == MessageIdentifiers.ID_NO_FREE_INCOMING_CONNECTIONS:
            print("The server is full.")
        elif packet.data[0] == MessageIdentifiers.ID_DISCONNECTION_NOTIFICATION:
            print("We have been disconnected.")
        elif packet.data[0] == MessageIdentifiers.ID_CONNECTION_LOST:
            print("Connection lost.")
        if packet.data[0] == MessageIdentifiers.ID_CONNECTION_REQUEST_ACCEPTED:
            print("Our connection request has been accepted.")
            host, port = packet.system_address
            client.send(
                bytes([ID_GAME_MESSAGE_1]) + b"Hello, World",
                PacketPriority.HIGH_PRIORITY,
                PacketReliability.RELIABLE_ORDERED,
                0,
                host,
                port,
            )
            client.shutdown(1)
            return

    assert False, "Connection timed out"


def test_echo():
    server = RakPeer()
    server.max_incoming_connections = 10
    server.startup()
    server_host, server_port = server.get_bound_address()

    client = RakPeer()
    client.startup()
    client.connect(server_host, server_port)

    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        futures = [executor.submit(run_server, server), executor.submit(run_client, client)]
        for future in futures:
            future.result()



import time

from raknet import RakPeer, MessageIdentifiers

ID_GAME_MESSAGE_1 = MessageIdentifiers.ID_USER_PACKET_ENUM + 1


def run(server_ver: int, client_ver: int):
    server = RakPeer()
    server.max_incoming_connections = 10
    server.startup(max_connections=10, protocol_version=server_ver)
    server_host, server_port = server.get_bound_address()

    client = RakPeer()
    client.startup(protocol_version=client_ver)
    client.connect(server_host, server_port)

    timeout = time.time() + 5
    while time.time() < timeout:
        packet = client.receive()
        if packet is None:
            continue

        if packet.data[0] == MessageIdentifiers.ID_CONNECTION_REQUEST_ACCEPTED:
            assert server_ver == client_ver
            return
        elif packet.data[0] == MessageIdentifiers.ID_INCOMPATIBLE_PROTOCOL_VERSION:
            assert server_ver != client_ver
            return

    assert False, "Timed out without receiving expected server response"


def test_compatible_1():
    run(6, 6)


def test_compatible_2():
    run(11, 11)


def test_incompatible_1():
    run(7, 8)


def test_incompatible_2():
    run(10, 9)



import os

from conan import ConanFile
from conan.tools.build import can_run
from conan.tools.cmake import cmake_layout, CMake


class TestPackageConan(ConanFile):
    settings = "os", "arch", "compiler", "build_type"
    generators = "CMakeDeps", "CMakeToolchain", "VirtualRunEnv"
    test_type = "explicit"

    def requirements(self):
        self.requires(self.tested_reference_str)

    def layout(self):
        cmake_layout(self)

    def build(self):
        cmake = CMake(self)
        cmake.configure()
        cmake.build()

    def test(self):
        if can_run(self):
            bin_path = os.path.join(self.cpp.build.bindir, "test_package")
            self.run(bin_path, env="conanrun")



import os

from conan import ConanFile
from conan.tools.cmake import CMake, CMakeDeps, CMakeToolchain, cmake_layout
from conan.tools.files import apply_conandata_patches, copy, export_conandata_patches, get

required_conan_version = ">=1.53.0"


class RakNetConan(ConanFile):
    name = "raknet"
    description = "RakNet is a cross platform, open source, C++ networking engine for game programmers."
    license = "BSD-2-Clause"
    url = "https://github.com/wu-vincent/raknet-python"
    homepage = "https://github.com/facebookarchive/RakNet"
    topics = ("raknet", "network", "game")
    package_type = "library"
    settings = "os", "arch", "compiler", "build_type"
    options = {
        "shared": [True, False],
        "fPIC": [True, False],
    }
    default_options = {
        "shared": False,
        "fPIC": True,
    }

    def export_sources(self):
        export_conandata_patches(self)

    def config_options(self):
        if self.settings.os == "Windows":
            del self.options.fPIC

    def configure(self):
        if self.options.shared:
            self.options.rm_safe("fPIC")

    def layout(self):
        cmake_layout(self, src_folder="src")

    def source(self):
        get(self, **self.conan_data["sources"][self.version], strip_root=True)

    def generate(self):
        tc = CMakeToolchain(self)
        tc.variables["RAKNET_ENABLE_SAMPLES"] = False
        tc.variables["RAKNET_ENABLE_DLL"] = self.options.shared
        tc.variables["RAKNET_ENABLE_STATIC"] = not self.options.shared
        tc.variables["RAKNET_GENERATE_INCLUDE_ONLY_DIR"] = True
        tc.generate()
        tc = CMakeDeps(self)
        tc.generate()

    def build(self):
        apply_conandata_patches(self)
        cmake = CMake(self)
        cmake.configure()
        cmake.build()

    def package(self):
        copy(self, "LICENSE", self.source_folder, os.path.join(self.package_folder, "licenses"))
        cmake = CMake(self)
        cmake.install()

    def package_info(self):
        if self.options.shared:
            self.cpp_info.libs = ["raknet"]
        else:
            self.cpp_info.libs = ["raknet-static"]

        self.cpp_info.set_property("cmake_file_name", "raknet")
        self.cpp_info.set_property("cmake_target_name", "raknet::raknet")

        if self.settings.os in ["Linux", "FreeBSD"]:
            self.cpp_info.system_libs.append("pthread")
        elif self.settings.os == "Windows":
            self.cpp_info.system_libs.append("ws2_32")

        # TODO: to remove in conan v2 once cmake_find_package_* generators removed
        self.cpp_info.names["cmake_find_package"] = "raknet"
        self.cpp_info.names["cmake_find_package_multi"] = "raknet"



# @formatter:off # fmt: off
import typing


class ConnectionAttemptError(RuntimeError):
    ...


class StartupError(RuntimeError):
    ...


class MessageIdentifiers:
    ID_ADVERTISE_SYSTEM: typing.ClassVar[int] = 29
    ID_ALREADY_CONNECTED: typing.ClassVar[int] = 18
    ID_CONNECTED_PING: typing.ClassVar[int] = 0
    ID_CONNECTED_PONG: typing.ClassVar[int] = 3
    ID_CONNECTION_ATTEMPT_FAILED: typing.ClassVar[int] = 17
    ID_CONNECTION_BANNED: typing.ClassVar[int] = 23
    ID_CONNECTION_LOST: typing.ClassVar[int] = 22
    ID_CONNECTION_REQUEST: typing.ClassVar[int] = 9
    ID_CONNECTION_REQUEST_ACCEPTED: typing.ClassVar[int] = 16
    ID_DETECT_LOST_CONNECTIONS: typing.ClassVar[int] = 4
    ID_DISCONNECTION_NOTIFICATION: typing.ClassVar[int] = 21
    ID_DOWNLOAD_PROGRESS: typing.ClassVar[int] = 30
    ID_INCOMPATIBLE_PROTOCOL_VERSION: typing.ClassVar[int] = 25
    ID_INVALID_PASSWORD: typing.ClassVar[int] = 24
    ID_IP_RECENTLY_CONNECTED: typing.ClassVar[int] = 26
    ID_NEW_INCOMING_CONNECTION: typing.ClassVar[int] = 19
    ID_NO_FREE_INCOMING_CONNECTIONS: typing.ClassVar[int] = 20
    ID_OPEN_CONNECTION_REPLY_1: typing.ClassVar[int] = 6
    ID_OPEN_CONNECTION_REPLY_2: typing.ClassVar[int] = 8
    ID_OPEN_CONNECTION_REQUEST_1: typing.ClassVar[int] = 5
    ID_OPEN_CONNECTION_REQUEST_2: typing.ClassVar[int] = 7
    ID_OUR_SYSTEM_REQUIRES_SECURITY: typing.ClassVar[int] = 11
    ID_OUT_OF_BAND_INTERNAL: typing.ClassVar[int] = 13
    ID_PUBLIC_KEY_MISMATCH: typing.ClassVar[int] = 12
    ID_REMOTE_SYSTEM_REQUIRES_PUBLIC_KEY: typing.ClassVar[int] = 10
    ID_SND_RECEIPT_ACKED: typing.ClassVar[int] = 14
    ID_SND_RECEIPT_LOSS: typing.ClassVar[int] = 15
    ID_TIMESTAMP: typing.ClassVar[int] = 27
    ID_UNCONNECTED_PING: typing.ClassVar[int] = 1
    ID_UNCONNECTED_PING_OPEN_CONNECTIONS: typing.ClassVar[int] = 2
    ID_UNCONNECTED_PONG: typing.ClassVar[int] = 28
    ID_USER_PACKET_ENUM: typing.ClassVar[int] = 134


class Packet:
    @property
    def data(self) -> bytes: ...
    @property
    def system_address(self) -> tuple[str, int]: ...


class PacketPriority:
    HIGH_PRIORITY: typing.ClassVar[PacketPriority]  # value = <PacketPriority.HIGH_PRIORITY: 1>
    IMMEDIATE_PRIORITY: typing.ClassVar[PacketPriority]  # value = <PacketPriority.IMMEDIATE_PRIORITY: 0>
    LOW_PRIORITY: typing.ClassVar[PacketPriority]  # value = <PacketPriority.LOW_PRIORITY: 3>
    MEDIUM_PRIORITY: typing.ClassVar[PacketPriority]  # value = <PacketPriority.MEDIUM_PRIORITY: 2>


class PacketReliability:
    RELIABLE: typing.ClassVar[PacketReliability]  # value = <PacketReliability.RELIABLE: 2>
    RELIABLE_ORDERED: typing.ClassVar[PacketReliability]  # value = <PacketReliability.RELIABLE_ORDERED: 3>
    RELIABLE_ORDERED_WITH_ACK_RECEIPT: typing.ClassVar[PacketReliability]  # value = <PacketReliability.RELIABLE_ORDERED_WITH_ACK_RECEIPT: 7>
    RELIABLE_SEQUENCED: typing.ClassVar[PacketReliability]  # value = <PacketReliability.RELIABLE_SEQUENCED: 4>
    RELIABLE_WITH_ACK_RECEIPT: typing.ClassVar[PacketReliability]  # value = <PacketReliability.RELIABLE_WITH_ACK_RECEIPT: 6>
    UNRELIABLE: typing.ClassVar[PacketReliability]  # value = <PacketReliability.UNRELIABLE: 0>
    UNRELIABLE_SEQUENCED: typing.ClassVar[PacketReliability]  # value = <PacketReliability.UNRELIABLE_SEQUENCED: 1>
    UNRELIABLE_WITH_ACK_RECEIPT: typing.ClassVar[PacketReliability]  # value = <PacketReliability.UNRELIABLE_WITH_ACK_RECEIPT: 5>


class RakPeer:
    max_incoming_connections: int
    def __init__(self) -> None:
        ...
    def connect(self, host: str, port: int, num_attempts: int = 6, attempt_interval_ms: int = 1000, timeout: int = 0) -> None:
        ...
    def get_bound_address(self, index: int = 0) -> tuple[str, int]:
        ...
    def receive(self) -> Packet:
        ...
    def send(self, data: bytes, priority: PacketPriority, reliability: PacketReliability, ordering_channel: int, host: str, port: int, force_receipt_num: int = 0) -> int:
        ...
    def shutdown(self, timeout_secs: float, ordering_channel: int = 0, disconnection_notification_priority: PacketPriority = ...) -> None:
        ...
    def startup(self, host: str = None, port: int = 0, max_connections: int = 1, protocol_version: int = 6, max_internal_ids: int = 10) -> None:
        ...
    @property
    def active(self) -> bool:
        ...
    @property
    def num_connections(self) -> int:
        ...

# @formatter:on # fmt: on


from conan import ConanFile
from conan.tools.cmake import CMakeToolchain, CMake, cmake_layout, CMakeDeps

__version__ = "0.2.0"


class RakNetPythonRecipe(ConanFile):
    name = "raknet-python"
    version = __version__
    package_type = "shared-library"

    # Optional metadata
    license = "BSD-3-Clause"
    url = "https://github.com/wu-vincent/raknet-python"
    homepage = "https://github.com/wu-vincent/raknet-python"
    description = "Python bindings for the RakNet network library"
    topics = ("python", "raknet", "network")

    # Binary configuration
    settings = "os", "compiler", "build_type", "arch"
    options = {}
    default_options = {"raknet/*:shared": False}

    exports_sources = "CMakeLists.txt", "src/*", "include/*"

    def requirements(self):
        self.requires("raknet/4.081")
        self.requires("pybind11/2.11.1")
        # self.test_requires("gtest/1.14.0")

    def layout(self):
        cmake_layout(self)

    def generate(self):
        deps = CMakeDeps(self)
        deps.generate()
        tc = CMakeToolchain(self)
        tc.generate()

    def build(self):
        cmake = CMake(self)
        cmake.configure()
        cmake.build()

    def package(self):
        cmake = CMake(self)
        cmake.install()

    def package_info(self):
        self.cpp_info.requires = ["raknet::raknet", "pybind11::headers"]



import pytest

from raknet import RakPeer, StartupError


def test_server_startup():
    server = RakPeer()
    server.startup(max_connections=10)


def test_client_startup():
    client = RakPeer()
    client.startup()


def test_duplicate_port():
    alice = RakPeer()
    alice.startup()
    _, port = alice.get_bound_address()
    with pytest.raises(StartupError):
        bob = RakPeer()
        bob.startup(port=port)


def test_negative_max_connections():
    with pytest.raises(StartupError):
        peer = RakPeer()
        peer.startup(max_connections=0)


# def test_bad_hostname():
#     with pytest.raises(StartupError):
#         peer = RakPeer()
#         peer.startup(host="bad-host-name")


def test_invalid_max_internal_ids():
    with pytest.raises(StartupError):
        peer = RakPeer()
        peer.startup(max_internal_ids=21)
    with pytest.raises(StartupError):
        peer = RakPeer()
        peer.startup(max_internal_ids=0)
